'
'	pb_sqlit.bas
'
'	DESCRIPTION
'	wrapper for sqlite.dll. sqlite.dll is the public domain sqlite sql
'	database engine from http://www.sqlite.org Please see that site for
'	availability of the sqlite.dll file.
'
'	COPYRIGHT
'	Copyright Don Dickinson, All rights reserved.
'	http://www.greatwebdivide.com
'	mailto:don@greatwebdivide.com
'
'	LICENSE
'	This code is free for all to use. No guarantees or warranties of any
'	kind are granted. You are using this free code at your own risk. Its
'	use implies that you accept this fact. You may redistribute it to others
'	as long as you are distributing the original file and leave this file
'	header in tact. This file is dependent on the following source files
'	made by Don Dickinson which are either under the same license or
'	are public domain:
'		pb_sqlit.inc 	-> inclean'd winapi declarations
'		pb_mem.bas 		-> memory handling
'		dd_link.bas 	-> linked lists
'	The linked list library is extensive. Only a few of the calls in it
'	are actually used. It would be possible to make the compilation smaller
'	by eliminating unused calls from it. If you do this, please rename the
'	file so it doesn't conflict with anyone using the original file.
'
'	COMPILING
'	This requires pbdll 6.1 or greater to compile as a dll.
'	It may also be compiled into your code with pbcc2.1 or greater.
'	There is a small demo program in this header. Just un-commennt the
'	%COMPILE_TEST_PROGRAM line below and compile with pbcc2.1 or greater.
'
#if not %def(%PB_SQLIT_BAS)
%PB_SQLIT_BAS = 1

'- NOTE: if you uncomment the line below
'			and compile with pbcc2.1 or greater
'			you will be compiling a small test
'			program that creates a database table
'			and puts some data into it.
'
%COMPILE_TEST_PROGRAM = 1

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'	FOR TEST PROGRAM ONLY (see %COMPILE_TEST_PROGRAM equate above
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#if %def(%COMPILE_TEST_PROGRAM)
#compile exe
#dim all
#endif
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#include "pb_sqlit.inc"			'inclean'd api
#include "pb_mem.bas"			'memory allocation
#include "dd_link.bas"			'linked lists

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'	FOR TEST PROGRAM ONLY
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#if %def(%COMPILE_TEST_PROGRAM)
#include "dd_stdout.bas"
#endif
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'	GLOBALS
'	needed to keep track of options and initialization status.
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
global g_sqlite_Init as long
global g_sqlite_BusyRetry as long
global g_sqlite_RetryWait as long

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'	CONSTANTS
'	first ones are defaults for the g_sqlite_* variables
'	the rest are return codes for the various sqlite functions
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
%PBSQLITE_DEFAULT_BUSY_RETRY = 10
%PBSQLITE_DEFAULT_RETRY_WAIT = 500

%SQLITE_ISO8859         =   1
%SQLITE_OK              =   0   '*/ Successful result */
%SQLITE_ERROR           =   1   '*/ SQL error or missing database */
%SQLITE_INTERNAL        =   2   '*/ An internal logic error in SQLite */
%SQLITE_PERM            =   3   '*/ Access permission denied */
%SQLITE_ABORT           =   4   '*/ Callback routine requested an abort */
%SQLITE_BUSY            =   5   '*/ The database file is locked */
%SQLITE_LOCKED          =   6   '*/ A table in the database is locked */
%SQLITE_NOMEM           =   7   '*/ A malloc() failed */
%SQLITE_READONLY        =   8   '*/ Attempt to write a readonly database */
%SQLITE_INTERRUPT       =   9   '*/ Operation terminated by sqlite_interrupt() */
%SQLITE_IOERR           =   10   '*/ Some kind of disk I/O error occurred */
%SQLITE_CORRUPT         =   11   '*/ The database disk image is malformed */
%SQLITE_NOTFOUND        =   12   '*/ (Internal Only) Table or record not found */
%SQLITE_FULL            =   13   '*/ Insertion failed because database is full */
%SQLITE_CANTOPEN        =   14   '*/ Unable to open the database file */
%SQLITE_PROTOCOL        =   15   '*/ Database lock protocol error */
%SQLITE_EMPTY           =   16   '*/ (Internal Only) Database table is empty */
%SQLITE_SCHEMA          =   17   '*/ The database schema changed */
%SQLITE_TOOBIG          =   18   '*/ Too much data for one row of a table */
%SQLITE_CONSTRAINT      =   19   '*/ Abort due to contraint violation */
%SQLITE_MISMATCH        =   20   '*/ Data type mismatch */
%SQLITE_MISUSE          =   21   '*/ Library used incorrectly */
%SQLITE_NOLFS           =   22   '*/ Uses OS features not supported on host */
%SQLITE_AUTH            =   23   '*/ Authorization denied */
%SQLITE_ROW             =   100  '*/ sqlite_step() has another row ready */
%SQLITE_DONE            =   101  '*/ sqlite_step() has finished executing */


'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'	STRUCTURES
'	This structure is used by this wrapper only. it is not part of sqlite.
'	A pointer to this structure is passed as the pExecute parameter in
'	various wrapper functions.
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
type INTERNAL_VMM_TYPE
	pVMM as long
	pFields as long
end type

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'	FUNCTION DECLARATIONS
'	these declare the sqlite function calls in sqlite.dll
'	note that they are all CDECL. if you do it without this, errors and
'	strange things will happen in your progrogram eventually.
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
declare function sqlite_open cdecl lib "sqlite.dll" alias "sqlite_open" _
		( zFilename as asciiz, byval iMode as long, pzErrMsg as asciiz ptr) as long
declare sub sqlite_close cdecl lib "sqlite.dll" alias "sqlite_close" _
		( byval hSQLite as long )

declare sub sqlite_freemem cdecl lib "sqlite.dll" alias "sqlite_freemem" _
		( byval pMem as long )
declare function sqlite_last_insert_rowid cdecl lib "sqlite.dll" alias "sqlite_last_insert_rowid" _
		( byval hSQLite as long ) as long

declare function sqlite_changes cdecl lib "sqlite.dll" alias "sqlite_changes" _
		( byval hSQLite as long ) as long
declare function sqlite_interrupt cdecl lib "sqlite.dll" alias "sqlite_interrupt" _
		( byval hSQLite as long ) as long
declare function sqlite_complete cdecl lib "sqlite.dll" alias "sqlite_complete" _
		( zSQL as asciiz ) as long
declare function sqlite_get_table cdecl lib "sqlite.dll" alias "sqlite_get_table" _
		( byval hSQLite as long, zSQL as asciiz, lpTable as long, iRow as long, _
		  iCol as long, lpErrMsg as long ) as long
declare function sqlite_free_table cdecl lib "sqlite.dll" alias "sqlite_free_table" _
		( byval lpTable as long ) as long

declare function sqlite_compile cdecl lib "sqlite.dll" alias "sqlite_compile" _
		( byval hSQLite as long, zSQL as asciiz, pzTail as asciiz ptr, _
		  ppVmm as long ptr, pzErr as asciiz ptr ) as long
declare function sqlite_finalize cdecl lib "sqlite.dll" alias "sqlite_finalize" _
		( byval pVmm as long, pzErr as asciiz ptr ) as long
declare function sqlite_step cdecl lib "sqlite.dll" alias "sqlite_step" _
		( byval pVmm as long, iCount as long, pazValue as long, _
		  pazColName as long ) as long

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'	sqliteInit
'
'	initializes the sqlite wrapper. call this exactly once at the beginning
'	of your app before other calls. Call the sister function sqliteDone
'	exactly once before finishing.
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
sub sqliteInit alias "sqliteInit" export

	if g_sqlite_Init then exit sub

	g_sqlite_Init = %true
	g_sqlite_BusyRetry = %PBSQLITE_DEFAULT_BUSY_RETRY
	g_sqlite_RetryWait = %PBSQLITE_DEFAULT_RETRY_WAIT

end sub

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'	sqliteDone
'
'	Cleans up the sqlite wrapper. call this exactly once before your
'	app exits.
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
sub sqliteDone alias "sqliteDone" () export

	if g_sqlite_Init = %false then exit sub
	g_sqlite_Init = %false

end sub

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'	sqliteUseDB
'
'	Opens a database file. It always opens in read-write mode unless the
'	file itself is readonly. Then it opens in Read-Only mode. this is the
'	current behavior of the sqlite engine. the sqlite_open "mode"
'	parameter is ignored.
'
'	Returns:
'		0 on failure and fills sErr with a text error message.
'		a non-zero handle (hDB parameters) to be passed to other engine calls.
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function sqliteUseDB alias "sqliteUseDB" _
		( byval sDBName as string, sErr as string ) export as long

	dim hDB as long
	dim pzError as asciiz ptr

	if g_sqlite_Init = %false then sqliteInit

	hDB = sqlite_open ( sDBName + $nul, 0, pzError )
	if hDB = 0 then

		if pzError then
			sErr = @pzError
			sqlite_freemem pzError
		else
			sErr = "Unknown error opening or creating " + sDBName
		end if
	else
		function = hDB
	end if

end function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'	sqliteCloseDB
'
'	Closes an sqlite database that was opened with sqliteUseDB
'	Pass it the hDB handle returned by the sqliteUseDB call.
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
sub sqliteCloseDB alias "sqliteCloseDB" ( byval hDB as long ) export

	if g_sqlite_Init = %false then sqliteInit
	sqlite_close hDB

end sub

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'	sqliteExecute
'
'	Executes an sql statement. Use its return value in the sqliteNextRow
'	and sqliteExecuteDone functions.
'
'	Returns:
'		0 		on failure and fills the sErr parameter with a description
'				of the error.
'		<> 0  on success. This is a pExecute handle to be passed to
'				subsequent calls to sqliteNextRow and sqliteExecuteDone
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function sqliteExecute alias "sqliteExecute" _
		( byval hDB as long, byval sSQL as string, _
		  sErr as string ) export as long

	dim bOk as long
	dim iResult as long
	dim pzTail as asciiz ptr
	dim ppVmm as long ptr
	dim iFoo as long
	dim pzError as asciiz ptr
	dim pResult as INTERNAL_VMM_TYPE PTR

	if g_sqlite_Init = %false then sqliteInit

	bOk = %true
	sErr = ""

	iFoo = 0
	ppVmm = varptr(iFoo)

	'- get the vm for processing sql statement
	iResult = sqlite_compile ( hDB, sSQL + $nul, pzTail, _
		  ppVmm, pzError )
	if iResult <> %SQLITE_OK then
		bOk = %false
		if pzError then
			sErr = "sqlite_compile failure: " + @pzError
		else
			sErr = "sqlite_compile_failure: error message unknown"
			sqlite_freemem pzError
			pzError = 0
		end if
		goto sqliteExecute_done
	end if

	'- error message? it probably shouldn't happen here, but
	'	this doesn't hurt.
	'
	if pzError then
		sqlite_freemem pzError
		pzError = 0
	end if

sqliteExecute_done:

	'- free the virtual machine if there was an error
	if (@ppVmm <> 0) and (bOk = %false) then
		iResult = sqlite_finalize(@ppVmm, pzError)
		if iResult <> %SQLITE_OK then
			if pzError then
				if sErr = "" then
					sErr = @pzError
				else
					sErr = sErr + ", " + @pzError
				end if
				sqlite_freemem pzError
				pzError = 0
			end if
		end if
	end if

	'- our result
	if bOk = %false then
		function = 0

	else
		pResult = getMem(sizeof(@pResult))
		@pResult.pFields = llAdd(0, "*root*", "*root*")
		@pResult.pVMM = @ppVmm

		function = pResult

	end if

end function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'	sqliteExecuteDone
'
'	Cleans up after a successful sqliteExecute command.
'	pass the return value of sqliteExecute* in the pExecute parameter
'	sErr contains the error message if this generates an error.
'
'	Returns:
'		%true 		on success and fills the sErr parameter
'		%false 		on failure
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function sqliteExecuteDone alias "sqliteExecuteDone" _
		( byval pExecute as long, sErr as string ) export as long

	dim bOk as long
	dim iResult as long
	dim pzError as asciiz ptr
	dim pInt as INTERNAL_VMM_TYPE PTR

	if g_sqlite_Init = %false then sqliteInit

	pInt = pExecute
	sErr = ""
	bOk = %true

	'- huh? this shouldn't happen
	if pInt = 0 then
		function = bOk
		exit function
	end if

	'- free the virtual machine.
	if @pInt.pVMM then
		iResult = sqlite_finalize(@pInt.pVMM, pzError)
		if iResult <> %SQLITE_OK then
			if pzError then
				sErr = "ERROR " + format$(iResult) + ": " + @pzError
				sqlite_freemem pzError
				bOk = %false
				pzError = 0
			end if
		end if
	end if

	'- free the field pointers
	if @pInt.pFields then llFree @pInt.pFields

	'- free the structure
	freeMem pInt

	function = bOk

end function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'	sqliteNextRow
'
'	Read the next row of data. You should call this function until
'	it returns %false (indicating failure or end of data)
'
'	Returns
'		%true 		if there is data available to read. if %true is
'						returned, you can read the data with sqliteField
'		%false		if there was an error or there is no more data.
'						sErr is set to "EOF" if there is no more data,
'						otherwise sErr will contain an error message.
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function sqliteNextRow alias "sqliteNextRow" _
		( byval pExecute as long, sErr as string ) as long

	dim i as long
	dim iCount as long
	dim iCols as long
	dim iResult as long
	dim pColData as long ptr
	dim pColName as long ptr
	dim pzName as asciiz ptr
	dim pzValue as asciiz ptr
	dim pInt as INTERNAL_VMM_TYPE PTR
	dim pzError as asciiz ptr

	if g_sqlite_Init = %false then sqliteInit
	sErr = ""
	pInt = pExecute

	iResult = sqlite_step(@pInt.pVMM, iCols, pColData, pColName)
	do
		if iResult = %SQLITE_BUSY then
			incr iCount
			if iCount > g_sqlite_BusyRetry then
				function = %false
				sErr = "Error. SQL Lite query engine busy and/or timed out."
				exit do
			end if
			apiSleep g_sqlite_RetryWait

		elseif iResult = %SQLITE_DONE then
			function = %false
			sErr = "EOF"
			exit do

		elseif iResult = %SQLITE_ROW then

			'- read the data into the linked list
			for i = 1 to iCols
				pzName = @pColName
				@pzName = trim$(lcase$(@pzName))
				pzValue = @pColData
				llUpdateByName @pInt.pFields, @pzName, @pzValue
				pColName = pColName + 4
				pColData = pColData + 4
			next i

			function = %true
			exit do

		else				'%SQLITE_ERROR or %SQLLIST_MISUSE
			sErr = "sqliteNextRow ERROR: " + format$(iResult)
			function = %false
			exit do

		end if
	loop

	if pzError then
		if sErr = "" then
			sErr = @pzError
		else
			sErr = sErr + ", " + @pzError
		end if
		sqlite_freemem pzError
	end if

end function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'	sqlField
'
'	Returns the string contents of a field in the row obtained by the
'	last call to sqliteNextRow
'
'	Returns:
'		the string contents of the field
'		an empty string if the field name is invalid.
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function sqliteField alias "sqliteField" _
		( byval pExecute as long, byval fieldName as string ) as string

	dim pInt as INTERNAL_VMM_TYPE PTR

	pInt = pExecute
	if pInt = 0 then
		function = ""
		exit function
	end if

	fieldName = trim$(lcase$(fieldName))
	function = llDataByName(@pInt.pFields, fieldName)

end function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'	sqliteExecuteNoReturn
'
'	Executes a SQL statement and just determines if it was successful. It
'	does not report an row data. Use this for INSERT, UPDATE, DELETE statements
'	where you won't expect a result set. It cleans up after itself, so there
'	is no need to call sqlliteExecuteDone
'
'	Returns
'		%true		on success
'		%false	on failure
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function sqliteExecuteNoReturn alias "sqliteExecuteNoReturn" _
		( byval hDB as long, byval sSQL as string, _
		  sErr as string ) export as long

	dim hSQL as long
	dim sFoo as string

	hSQL = sqliteExecute(hDB, sSQL, sErr)
	if hSQL = 0 then
		function = %false
		exit function
	end if

	do until sqliteNextRow(hSQL, sErr) = %false
	loop

	if ucase$(sERR) = "EOF" then
		function = sqliteExecuteDone(hSQL, sErr)
	else
		'- we want to preserve sErr, so pass sFoo
		'	here in order to keep it.
		'
		sqliteExecuteDone hSQL, sFoo
		function = %false
	end if

end function

#if %def(%COMPILE_TEST_PROGRAM)
$TEST_FILE = ".\test_demo_db.sqlite"
$SQL_TABLE_CREATE = "CREATE TABLE t_test (nm char(40), count_int int)"

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'	pbmain
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function pbmain

	dim hDB as long
	dim hSQL as long
	dim i as long
	dim iCount as long
	dim sSQL as string
	dim sErr as string

	'- Always the first call your app should make.
	sqliteInit

	'- Delete the test database if it exists. Then Open it.
	if dir$($TEST_FILE) <> "" then
		kill $TEST_FILE
	end if

	hDB = sqliteUseDB($TEST_FILE, sErr)
	if hDB = 0 then
		stdout "error: " + sErr
		goto pbmain_done
	end if

	'- Execute the sql statement to create a database
	stdout "OK about to execute"
	if sqliteExecuteNoReturn(hDB, $SQL_TABLE_CREATE, sErr) = %false then
		stdout "ERROR creating table"
		stdout sErr
		goto pbmain_done
	end if
	stdout "Table created."

	'- Begin a transaction. If you don't wrap your multiple transactions
	'	inside of a transaction, they will go much, much slower.
	'
	if sqliteExecuteNoReturn(hDB, "BEGIN TRANSACTION", sErr) = %false then
		stdout "ERROR beginning transaction"
		stdout sErr
		goto pbmain_done
	end if

	'- Insert 100 records
	for i = 1 to 100
		if i mod 100 = 0 then
			stdoutCL "Inserted " + format$(i)
		end if
		sSQL = "INSERT INTO t_test (nm,count_int) " + _
				"VALUES ('Name " + format$(i) + "', " + format$(i) + ")"
		if sqliteExecuteNoReturn(hDB, sSQL, sErr) = %false then
			stdout "ERROR inserting record: " + sErr
			goto pbmain_done
		end if

	next i

	'- Commit them.
	stdout "END TRANSACTION"
	if sqliteExecuteNoReturn(hDB, "END TRANSACTION", sErr) = %false then
		stdout "ERROR ending transaction"
		stdout sErr
		goto pbmain_done
	end if

	stdout
	stdout "Done inserting"

	'- After we finish inserting, then read back the count.
	'	notice i loop through the records. this isn't probably
	'	necessary because only one record will be returned. That
	'	record will have only one field: the count.
	'
	sSQL = "SELECT count(*) AS icount FROM t_test"
	hSQL = sqliteExecute(hDB, sSQL, sErr)

	if hSQL = 0 then
		stdout "error selecting count: " + sErr
	else
		do until sqliteNextRow(hSQL, sErr) = %false
			stdout "count = " + sqliteField(hSQL, "icount")
		loop
	end if

	'- After we're done looping, we have to cleanup the
	'	execution of the sql or we get a memory leek.
	'
	if sqliteExecuteDone(hSQL, sERR) = %false then
		stdout "ERROR finishing sql: " + sErr
	end if

	stdout "Press any key to read back the records (esc to quit)"
	if waitkey$ = $esc then goto pbmain_done

	'- Select all of the records
	sSQL = "SELECT * FROM t_test ORDER BY nm"
	hSQL = sqliteExecute(hDB, sSQL, sErr)
	if hSQL = 0 then
		stdout "Error executing SQL"
		stdout sSQL
		goto pbmain_done
	end if

	'- loop through them
	cls
	iCount = 0
	do until sqliteNextRow(hSQL, sErr) = %false
		incr iCount
		if iCount mod 23 = 0 then
			stdout "Press any key to continue"
			waitkey$
			cls
		end if
		stdout sqliteField(hSQL, "count_int") + ": " + sqliteField(hSQL, "nm")
	loop

	'- cleanup
	if sqliteExecuteDone(hSQL, sErr) = %false then
		stdout "Error closing sql statement: " + sErr
	end if

pbmain_done:

	if hDB then
		stdout "Closing database"
		sqliteCloseDB hDB
	end if

	stdout "DONE"
	waitkey$

end function
#endif
#endif
